import java.util.concurrent.ConcurrentHashMap;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

final int LEN = 100;
final int MIN = 1;
final int MAX = 100;
final int OFFSET = 10;
int swapCompareCost = 0;
boolean showTip = true;

SortingAlgorithm[] sorters = null;

void setup() {
  size(640, 480);
  frameRate(1000);
  colorMode(HSB, 360, 100, 100);
  ellipseMode(CENTER);
  textAlign(CENTER, CENTER);
  textFont(createFont("Liberation Mono", 12));

  resetSorters(newRandomArray());
}

void draw() {
  background(0);

  for (SortingAlgorithm sorter : sorters) {
    sorter.step();
  }

  stroke(#000000);
  textSize(12);
  drawAlgorithm(sorters[0], 0, 0, width / 2, height / 2);
  drawAlgorithm(sorters[1], width / 2, 0, width / 2, height / 2);
  drawAlgorithm(sorters[2], 0, height / 2, width / 2, height / 2);
  drawAlgorithm(sorters[3], width / 2, height / 2, width / 2, height / 2);
  
  if (!showTip) return;
  
  fill(#AA000000);
  noStroke();
  rect(0, 0, width, height);
  fill(#AA000000);
  stroke(#FFFFFF);
  rect(50, 50, width - 100, height - 100);
  
  fill(#FFFFFF);
  textSize(12);
  text("Comparison of InsertionSort, SelectionSort, BubbleSort & QuickSort\n\nTitle of Diagrams:\nNumber of (C)omparisons, Number of (S)waps, Runtime in (s)econds\n\nKeyboard Shortcuts [KEY]\n\nCost of Comparison/Swap Operation:\n[0] 0ms [1] 10ms [2] 25ms [3] 50ms [4] 100ms\n\nGenerate new data array:\n[n] random [s] sorted [r] reverse [a] almost sorted\n\nShow this help / tip:\n[t] show [h] hide\n\nStart simulation: [SPACE]\nClose program: [ESCAPE] or [q]", width / 2, height / 2);
}

void drawAlgorithm(SortingAlgorithm sorter, int x, int y, int width, int height) {
  if (sorter == null || sorter.data == null) return;

  fill(#FFFFFF);
  text(sorter.stat(), x + width / 2, y + 5);

  float barWidth = (width - 2.0 * OFFSET) / sorter.data.length;
  float barHeight = height - 2.0 * OFFSET - barWidth;
  float colorFactor = 360 / MAX;
  float xPos = OFFSET;

  for (int i = 0; i < sorter.data.length; i++) {
    fill(color(colorFactor * sorter.data[i], 100, 100));
    rect(x + xPos, y + height - OFFSET - barWidth, barWidth, -(barHeight / MAX) * sorter.data[i]);

    xPos += barWidth;
  }

  float radius = Math.max(5, barWidth / 2);

  for (java.util.Map.Entry<String, Integer> entry : sorter.markerIndices.entrySet()) {
    String markerName = entry.getKey();
    int markerIdx = entry.getValue();

    if (!sorter.markerColors.containsKey(markerName)) continue;

    color markerColor = sorter.markerColors.get(markerName);

    fill(markerColor);
    ellipse(x + OFFSET + barWidth * markerIdx + radius, y + height - OFFSET - radius, radius, radius);
  }

  if (sorter.compareIdxA < 0 || sorter.compareIdxB < 0) return;

  if (sorter.compareOrSwap) {
    fill(color(#FF0000));

    ellipse(x + OFFSET + barWidth * sorter.compareIdxA + radius, y + height - OFFSET, radius, radius);
    ellipse(x + OFFSET + barWidth * sorter.compareIdxB + radius, y + height - OFFSET, radius, radius);
  } else {
    fill(color(#00FF00));

    ellipse(x + OFFSET + barWidth * sorter.compareIdxA + radius, y + height - OFFSET, radius, radius);
    ellipse(x + OFFSET + barWidth * sorter.compareIdxB + radius, y + height - OFFSET, radius, radius);
  }
}

void resetSorters(int[] data) {
  boolean allFinished = true;

  if (sorters != null) {
    for (SortingAlgorithm sorter : sorters) {
      allFinished &= sorter.finished;
    }
  }

  if (!allFinished) return;

  sorters = new SortingAlgorithm[] {
    new SelectionSort(),
    new InsertionSort(),
    new BubbleSort(),
    new QuickSort()
  };

  for (SortingAlgorithm sorter : sorters) {
    sorter.setData(data.clone());
  }
}

void startSorters() {
  boolean allFinished = true;
  boolean allStarted = true;
  boolean allStopped = true;

  for (SortingAlgorithm sorter : sorters) {
    allFinished &= sorter.finished;
    allStarted &= sorter.started;
    allStopped &= sorter.stopped;
  }
  
  if (!allFinished || allStarted) {
    if (allStopped) {
      showTip = true;
    }
    
    return;
  }

  for (SortingAlgorithm sorter : sorters) {
    sorter.start();
  }
}

void keyPressed() {
  showTip = false;
  
  switch (key) {
  case '0':
    swapCompareCost = 0;
    break;
  case '1':
    swapCompareCost = 10;
    break;
  case '2':
    swapCompareCost = 25;
    break;
  case '3':
    swapCompareCost = 50;
    break;
  case '4':
    swapCompareCost = 100;
    break;
  case 'n':
    resetSorters(newRandomArray());
    break;
  case 'a':
    int[] a = sort(newRandomArray());

    for (int i = 0; i < a.length / 10; i++) {
      int idxA = (int)random(a.length);
      int idxB = (int)random(a.length);

      int tmp = a[idxA];
      a[idxA] = a[idxB];
      a[idxB] = tmp;
    }
    resetSorters(a);
    break;
  case 's':
    resetSorters(sort(newRandomArray()));
    break;
  case 'r':
    resetSorters(reverse(sort(newRandomArray())));
    break;
  case ' ':
    startSorters();
    break;
  case 't':
    showTip = true;
    break;
  case 'h':
    showTip = false;
    break;
  case 'q':
    exit();
    break;
  }
}

int[] newRandomArray() {
  int[] a = new int[LEN];

  for (int i = 0; i < a.length; i++) {
    a[i] = (int)Math.floor(Math.random() * (MAX - MIN) + MIN);
  }

  return a;
}
