abstract class SortingAlgorithm extends Thread {
  public int[] data;
  private boolean finished = true;
  private volatile boolean threadShouldWait = true;
  private int comparisons = 0;
  private int swaps = 0;
  private int startTime = -1;
  private int stopTime = -1;
  private boolean started = false;
  private boolean stopped = false;
  ConcurrentHashMap<String, Integer> markerColors = new ConcurrentHashMap<>();
  ConcurrentHashMap<String, Integer> markerIndices = new ConcurrentHashMap<>();
  int compareIdxA = -1;
  int compareIdxB = -1;
  boolean compareOrSwap = false;
  
  public void setData(int[] data) {
    this.data = data;
  }
  
  public abstract void sort();
  
  public void run() {
    clearMarkers();
    started = true;
    finished = false;
    comparisons = 0;
    swaps = 0;
    startTime = millis();
    stopTime = -1;
    this.sort();    
    finished = true;  
    stopTime = millis();
    stopped = true;
    clearMarkers();
  }
  
  public String stat() {
    int stop = stopTime == -1 ? millis() : stopTime;
    float time = startTime == -1 ? 0 : stop - startTime;
    
    time = round(time / 100) / 10.0;
    
    return this.getClass().getName().substring(getClass().getName().indexOf("$") + 1) + ": " + this.comparisons + "C / " + this.swaps + "S" + " / " + time + "s";
  }

  public void clearMarkers() {
    compareIdxA = -1;
    compareIdxB = -1;
    markerColors.clear();
    markerIndices.clear();
  }
  
  public boolean isFinished() {
    return finished;
  }
  
  public void step() {
    this.threadShouldWait = false;
  }
  
  private void sleepFor(int milliSeconds) {
    try {
      Thread.sleep(milliSeconds);
    } catch (InterruptedException e) {
    }
  }
  
  private void waitForNextStep() {
    while(this.threadShouldWait == true) {
      sleepFor(1);
    }
    
    this.threadShouldWait = true;
  }
  
  public void marker(String name, color col) {
    markerColors.put(name, col);
  }
  
  public void mark(String name, int idx) {
    waitForNextStep();
        
    markerIndices.put(name, idx);
  }
  
  public void unmark(String name) {
    waitForNextStep();
        
    markerIndices.remove(name);
  }
  
  public int compare(int idxA, int idxB) {
    if (idxA == idxB) return 0;
    
    waitForNextStep();
    
    compareIdxA = idxA;
    compareIdxB = idxB;
    compareOrSwap = false;
        
    sleepFor(swapCompareCost);
    
    comparisons++;
        
    return this.data[idxA] - this.data[idxB];
  }
  
  public void swap(int idxA, int idxB) {
    if (idxA == idxB) return;
    
    waitForNextStep();
    
    compareIdxA = idxA;
    compareIdxB = idxB;
    compareOrSwap = true;
        
    sleepFor(swapCompareCost);
    
    swaps++;
    
    int tmp = this.data[idxA];
    this.data[idxA] = this.data[idxB];
    this.data[idxB] = tmp;
  }
}
